%{
This script calculates the current distribution and ac loss in HTS tapes
located next to a point current source whilst carrying ac transport current.

Reference paper:
C. C. T. Chow, F. Grilli, and K. T. Chau, "Numerical modelling of HTS tapes 
under arbitrary external field and transport current via integral method: 
review and application to electrical machines," Supercond. Sci. Technol., 
2023. [Online]. Available: http://iopscience.iop.org/article/10.1088/1361-6668/ace701.

Author: Calvin C. T. Chow (ctcchow@eee.hku.hk)

Version history:
v1.0: October 2023 First published version. Use Matlab r2021b or newer

Code based on script "strip.m" by Dr. Simon Otten version dated 14 June 2019 (available at
https://www.htsmodelling.com/?wpdmpro=simonotten)

This program is free software: you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by the Free Software 
Foundation, either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT 
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with 
this program. If not, see <https://www.gnu.org/licenses/>.
%}


%% Setting
clear
mu0 = 4*pi*1e-7;

% tape properties
W = 0.004;          % tape width [m]
d = 1e-6;           % tape thickness [m]
Jc = 4e10*d;        % critical sheet current (critical current per unit width) [A/m]
Ic = Jc*W;          % critical current [A]
Ec = 1e-4;          % electric field at J=Jc [V/m]
n = 35;             % n-value
ny = 80;            % number of elements per tape for numerical calculation. 

%% coil positions settings
% coil is on 2D plane, width of the tape runs along y axis, thickness of
% tape runs along x axis
% Coil is a N_coil by 4 matrix. Each row represents a coil. The three
% columns are 
% 1) the x positions of the leftmost tape (smallest x value) of the coil [m]
% 2) the x position of the rightmost tape (largest x value) of the coil [m]
% 3) the y position of the bottom (smallest y value) of the coil. [m]
% 4) the number of turns in the coil [no unit]
% tapes are 1D objects with 0 distance along x and length W along y
Coil = [-0.001425, 0.001425, -0.0042, 10;
    -0.001425, 0.001425, 0.0002, 10] ;


    
%% applied current settings
I0 = 120;       % amplitude [A]
I_freq = 10;    % frequency [Hz]
alpha = 0;
S = 1;
phi = 0;

Iset = @(t) S*I0*sin(2*pi*I_freq*t+phi);

gamma = 100;        % feedback constant for voltage source [V/m/A]

%% external magnetic field settings
B_freq = 20;       % frequency [Hz]
B0 = 0.1;          % amplitude [T]
A_type = 7;

if A_type ==1
    % uniform field in the x direction
    Bext = @(t) B0*sin(2*pi*B_freq*t);
    Bdot = @(t) 2*pi*B_freq*B0*cos(2*pi*B_freq*t); % time derivative
    dAextdt_input = @(t,y,x) -y.*Bdot(t);
elseif A_type ==7
    % field created by point current source at (-0.01,0)
    dAextdt_input = @(t,y,x) -4*pi*1e-7*30000*2*pi*B_freq.*cos(2*pi*B_freq*t).*log((x-0.01).^2+y.^2)*0.5/2/pi;
end



%% time vector settings
N_step = 2000;         
period = 2*max(1/I_freq*(I0~=0), 1/B_freq*(B0~=0)); % 2 periods

dt = period/(N_step);  % time step [s]
t = 0:dt:period;       

fprintf('I0,I_freq,B0,B_freq,A_type,N_step,ny = %.3g,%.3g,%.3g,%.3g,%i,%i,%i \n', I0,I_freq,B0,B_freq,A_type,N_step,ny) %print up to 3sf

%% plot current
I_vec = Iset(t);
figure(1)
plot(t,I_vec)

dI_vec = gradient(I_vec,dt);
figure(2)
plot(t,dI_vec)

%%% END OF SETTINGS %%%

%% mesh coordinates of each element
% want the x value, center y value, lower y value and upper y value, and width for each
% element
y_lower_vec = [];
y_upper_vec = [];
y_middle_vec = [];
x_vec = [];
w_vec = [];

N_t = sum(Coil(:,4)); % total number of turns
N = ny * N_t; % total number of elements
L = zeros(N_t, N); % Lij = 1 if element j belongs to turn i, =0 otherwise
turn_num = 1; % counter of turns

% define lower and upper boundaries of the elements
for coil_num = 1:size(Coil,1) %for each coil
    x_smallest = Coil(coil_num,1);
    x_largest = Coil(coil_num,2);
    y_bottom = Coil(coil_num,3);
    num_turn = Coil(coil_num,4);

    x_tapes = linspace(x_smallest,x_largest,num_turn);
    x_vec = [x_vec, repelem(x_tapes,ny)];

    for tape_num = 1:Coil(coil_num,4) % for each turn 
        
        % y_vector is created for each coil. Between two consecutive numbers in
        % y_vector is an element 
        y_vector = linspace(y_bottom,y_bottom+W,ny+1); % option 1: linear spacing
        % y_vector = y_bottom + W/2 - W/2*cos((0:ny)/ny*pi); % option 2: smaller elements near edges (better for low amplitudes)
        
    
        y_lower = y_vector(1:end-1);     % lower boundaries of each element
        y_upper = y_vector(2:end);       % upper boundaries
        y_middle = (y_lower+y_upper)/2;  % middle point
        w = y_upper - y_lower;           % element width
        
        % add to existing vectors
        y_lower_vec = [y_lower_vec, y_lower];
        y_upper_vec = [y_upper_vec, y_upper];
        y_middle_vec = [y_middle_vec, y_middle];
        w_vec = [w_vec, w];
        
        % update L
        L(turn_num,((turn_num-1)*ny+1):(turn_num*ny)) = ones(1,ny);
        turn_num = turn_num+1;

    end 
end


%% calculate the K matrix (used as "mass matrix" for the ode solver)
K = zeros(N);
for i = 1:N
    yi = y_middle_vec(i);
    xi = x_vec(i);
    for j = 1:N
        aj = y_lower_vec(j);
        bj = y_upper_vec(j);
        xj = x_vec(j);
        if xi == xj
            K(i,j) = (yi-bj)*(log(abs(yi-bj))-1) -(yi-aj)*(log(abs(yi-aj))-1);
        else
            K(i,j) = (yi-bj)*(0.5*log((yi-bj)^2 +(xi-xj)^2)-1) +(xi-xj)*atan((yi-bj)/(xi-xj)) ...
                -(yi-aj)*(0.5*log((yi-aj)^2 +(xi-xj)^2)-1) -(xi-xj)*atan((yi-aj)/(xi-xj));
        end
    end
end

%% the differential equation
% electric field function and derivative
% J and Jc are sheet current density and sheet critical current density
E = @(J) Ec*sign(J).*abs(J./Jc).^n;
dEdJ = @(J) Ec*(n/Jc).*abs(J./Jc).^(n-1);

% initial current distribution
J0 = zeros(N,1); %J will be column vector

% gradient of the electric potential (voltage source)
dphi = @(t,J) gamma*(L'*L*(w_vec'.*J)-Iset(t));


% time derivative of Aext at each point
dAextdt = @(t,y,x) dAextdt_input(t,y_middle_vec', x_vec');


% right hand side of differential equation
fun = @(t,J) -(2*pi/mu0) *(E(J) +dAextdt(t) +dphi(t,J));
% Jacobian matrix
jac = @(t,J) -(2*pi/mu0) *(diag(dEdJ(J))+ gamma*(bsxfun(@times,w_vec,L'*L)));

% numerical integration
options = odeset();
options.AbsTol = 1e-3*Jc;
options.Jacobian = jac; 
options.Mass = K;
options.MStateDependence = 'none';
options.RelTol = 1e-6;
options.Stats = 'on';
options.Vectorized = 'on';

tic;
[T,J] = ode15s(fun,t,J0,options);
elapsed_time = toc

% T is a N_step+1 by 1 column vector, storing the time at each time step
% the result J is a N_step+1 by N matrix, each row represent the J at
% each of the N elements at a specific time

%% POST-PROCESSING 

% calculate the dissipated power at each time step
P = J.*E(J)*w_vec';
% calculate the AC loss per cycle by integrating the dissipated power 
% over the last cycle (two cycles are simulated)
Q = trapz(t(end-N_step/2:end),P(end-N_step/2:end));
dt = t(end)-t(end-N_step/2); %length of time integrated
Q = Q/dt;
disp(['Q = ',num2str(Q),' W/m']);


% calculate total current in each turn at each time step
I = L*(bsxfun(@times,w_vec,J))'; % I is N_t by N_step+1 matrix. 
I = I';
largest_difference = max(max(abs((I-Iset(t')))));
disp(['Largest difference between Iset and I is ',num2str(largest_difference), ' A'])


% plot power as function of time
figure(201)
clf
plot(T,P)
title(['Average Power in second half of cycle = ',num2str(Q),' W/m'])
xlabel('t [s]')
ylabel('P [W/m]')
grid

save("P", "P")
formatOut=30; %datetime format
timestring = datestr(datetime('now'), formatOut);
filename = strcat('./results/',timestring);
save(filename)