"""
This program can be used with ACloss_class.py to calculate the current distribution
of REBCO tapes under non-uniform external field (created by point current source) 
whilst carrying transport current. As an example, loss in 20 tapes is calculated below.

Reference paper:
C. C. T. Chow, F. Grilli, and K. T. Chau, "Numerical modelling of HTS tapes 
under arbitrary external field and transport current via integral method: 
review and application to electrical machines," Supercond. Sci. Technol., 
2023. [Online]. Available: http://iopscience.iop.org/article/10.1088/1361-6668/ace701.

Author: Calvin C. T. Chow (ctcchow@eee.hku.hk)

Version history:
v1.0: October 2023 First published version. In Python 3.8.

Based on code by Dr. Lingfeng Lai dated 2020/10/30 available at https://www.htsmodelling.com/?wpdmpro=j_model

This program is free software: you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by the Free Software 
Foundation, either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT 
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with 
this program. If not, see <https://www.gnu.org/licenses/>.

"""

from ACloss_class import YBCOtape, AClossFT2D

import numpy as np
import matplotlib.pyplot as plt
import time
from decimal import Decimal
import platform

computer_info = platform.processor() #check current computing platform for reporting purposes

### 1. Set the geometry of problem
# Define the YBCO tape. Use default parameters in the class file.
tape = YBCOtape()

# Set solver
solver = AClossFT2D(tape)

# Add coils 
x_offset = 0 #unit: meter. 
xleft = -(3e-3-150e-6)/9*4.5 +x_offset
xright = (3e-3-150e-6)/9*4.5 +x_offset
solver.add_coil(xleft, xright, -0.0042, 10)
solver.add_coil(xleft, xright, 0.0002, 10)


### 2. Set external conditions
## Set the transpose current
# Amplitude and frequency of transpose current
Imax = 120 #[A]
Ifreq = 10 #[Hz]
solver.set_current(Imax, Ifreq)

# set external magnetic vector potential (A)
Bmax = 0.1 #[T] # only relevant for A_typ =1. For A_type=7, the amplitude is defined in the ACloss_class file and is not adjustable here
Bfreq = 20 #[Hz]
A_type = 7 #  =1 for uniform external field, =7 for non-uniform external field generated by a point source
solver.set_A(Bmax, Bfreq,A_type)



### 3. Set solver parameters
# Divide each tape into ny elements along the y direction equally.
ny=80
solver.mesh(ny) 
# Set the simulated times. 
# Simulate over two periods (period being the longer period of I or B variation)
# Results (instantaneous loss, current density etc.) will be stored at exact time instants specified below
save_time = np.linspace(0, 2*max(1/Ifreq*(Imax!=0), 1/Bfreq*(Bmax!=0)), 2001, endpoint=True)

# minimum and maximum time steps
timestep_min = 1e-10
timestep_max = 1e-6

solver.set_time(save_time, timestep_min, timestep_max, 1)


### 4. Solve
t = time.time()
solver.run()
elapsed = time.time() - t
print('elapsed time (s)=', elapsed)


#save whatever result, may not have converged
save_result = 1

if save_result==1:
    timestr = time.asctime(time.localtime(time.time())).replace(":","")
    filename = './results/' + timestr + '.pkl'
    solver.save_result(filename)
    filename_computer = './results/' + timestr + 'computer.txt'


#plot results at specified frames
frames = [101,201] #enter in brackets the time_point_no
for frame in frames:
     solver.plot_resultj(frame) #normalized current density of all turns


#calculate the average ac loss for the second period simulated
power_avg, runtime_array,loss_array = solver.post_cal_power(1001,2000) #enter indices in save_time corresponding to the second half of the time simulated
np.savetxt('python_runtime_array.txt', runtime_array, delimiter=',')
np.savetxt('python_loss_array.txt', loss_array, delimiter=',')

# save results
if save_result==1:
    solver.save_result(filename)
    with open(filename_computer, 'w') as f:
        f.write(computer_info+'. time_taken = '+str(elapsed))




