"""
This program can be used with main.py to calculate the current distribution
of REBCO tapes under non-uniform external field (imported from external file) 
whilst carrying transport current.

Reference paper:
C. C. T. Chow, F. Grilli, and K. T. Chau, "Numerical modelling of HTS tapes 
under arbitrary external field and transport current via integral method: 
review and application to electrical machines," Supercond. Sci. Technol., 
2023. [Online]. Available: http://iopscience.iop.org/article/10.1088/1361-6668/ace701.

Author: Calvin C. T. Chow (ctcchow@eee.hku.hk)

Version history:
v1.0: October 2023 First published version. In Python 3.8.

Based on code by Dr. Lingfeng Lai dated 2020/10/30 available at https://www.htsmodelling.com/?wpdmpro=j_model

This program is free software: you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by the Free Software 
Foundation, either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT 
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with 
this program. If not, see <https://www.gnu.org/licenses/>.
"""


import numpy as np
import matplotlib.pyplot as plt
from matplotlib import cm
import time
import pickle
from scipy.interpolate import RegularGridInterpolator




class YBCOtape:
    """
    class for YBCO tape
    """
    def __init__(self, a=4e-3, b=1e-6, nvalue=35, jc0=4e10, B0=0.04265, k=0.29515, alpha=0.7):
        """
        define the YBCO tape
        :param a: tape width [m]
        :param b: tape thickness [m]
        :param nvalue: n value 
        :param jc0: critical current density [Am^-2]
        :param B0: parameter for Kim model
        :param k:  parameter for Kim model
        :param alpha:  parameter for Kim model
        """
        self.a = a
        self.b = b
        self.nvalue = nvalue
        self.jc = np.array(jc0)
        self.ic = jc0 * a * b
        self.B0 = B0
        self.k = k
        self.alpha = alpha


class Coil:
    """
    class for the YBCO coil/stack
    """
    def __init__(self, rin=0.1, rout=0.2, ymin=0.1, count_turns=10):
        """
        define the geometry of the coil/stack
        :param rin:  inner diameter [m]
        :param rout:  outer diameter [m]
        :param ymin:  the bottom position of the coil (bottom y position) [m]
        :param count_turns:  total number of turns of the coil
        """
        self.rin = rin
        self.rout = rout
        self.ymin = ymin
        self.count_turns = count_turns


class AClossFT2D:
    """
    class to solve the flux penetration process
    """
    def __init__(self, tape_type=YBCOtape()):
        """
        define the ReBCO tape used
        :param tape_type: (YBCOtape) YBCO tape defined
        """
        self.tape_type = tape_type
        self.coils = []
        self.save_time = []
        self.save_time_step = []
        self.j_limit = 1.1
        self.timestep_type = 0
        self.timestep_min = 1e-5
        self.timestep_max = 1e-4
        self.time_point_no = 0
        self.dta = []
        self.time_array = []
        self.power2 = [] #power2 is a matrix
        self.ny = 0
        self.dy = 0
        self.dx = 0
        self.count_turns = 0
        self.posi_x = np.empty(0)
        self.posi_y = np.empty(0)
        self.count_ele = 0
        self.Bax = 0
        self.Aa_fre = 0
        self.Aa_phi = 0
        self.A_type = 0
        self.Ia = 0
        self.Ia_fre = 0
        self.Ia_phi = 0
        self.jt = []
        self.dj = []
        self.jc = []
        self.L = None
        self.Qij_inv = None
        self.M = None
        self.LM_inv = None
        self.step = 0

        #parameters of ramp functions
        self.rm1_offset = 0.1;
        self.rm1_b = 0.001; #para
        self.rm2_offset = 1.8
        self.rm2_k = 5

    ## ramp functions and their derivatives        
    def _rm1(self, x):
        self.rm1_offset = 0.1;
        return ((x-self.rm1_offset)/2*(1+(x-self.rm1_offset)/(np.sqrt((x-self.rm1_offset)**2+self.rm1_b))) \
                     - (x-self.rm1_offset-1)/2*(1+(x-1-self.rm1_offset)/(np.sqrt((x-1-self.rm1_offset)**2+self.rm1_b))) ); # ramp function
    def _drm1(self, x):
        #derivative of rm1 wrt x
        ans =  (x-self.rm1_offset)/(np.sqrt((x-self.rm1_offset)**2+self.rm1_b)) \
            -0.5*(x-self.rm1_offset)**3*((x-self.rm1_offset)**2+self.rm1_b)**-1.5 \
            -(x-1-self.rm1_offset)/(np.sqrt((x-1-self.rm1_offset)**2+self.rm1_b))\
            +0.5*(x-1-self.rm1_offset)**3*((x-self.rm1_offset-1)**2+self.rm1_b)**-1.5 
        return ans
    
    def _rm2(self,x):
        #logistic function
        return 1/(1+np.exp(-self.rm2_k*(x-self.rm2_offset)))
    def _drm2(self,x):
        return self.rm2_k*self._rm2(x)*(1-self._rm2(x))
    
    # ramp function for external field
    def _rm_A(self,x):
        if self.rm_type_A ==1:
            return self._rm1(x)
        elif self.rm_type_A ==2:
            return self._rm2(x) 
        elif self.rm_type_A ==0:
            return 1    
    def _drm_A(self,x):
        if self.rm_type_A ==1:
            return self._drm1(x)
        elif self.rm_type_A ==2:
            return self._drm2(x)
        elif self.rm_type_A ==0:
            return 0
    
    # ramp function for transport current    
    def _rm_I(self,x):
            if self.rm_type_I ==1:
                return self._rm1(x)
            elif self.rm_type_I ==2:
                return self._rm2(x)
            elif self.rm_type_I ==0:
                return 1
    def _drm_I(self,x):
        if self.rm_type_I ==1:
            return self._drm1(x)
        elif self.rm_type_I ==2:
            return self._drm2(x)
        elif self.rm_type_I ==0:
            return 0

    def add_coil(self, rin=0.1, rout=0.2, ymin=0.0, count_turns=10):
        """
        adding a coil to the geometry of the problem
        :param rin:  inner diameter [m]
        :param rout:  outer diameter [m]
        :param ymin:  the bottom position of the coil (bottom y position) [m]
        :param count_turns:  total number of turns of the coil
        :return: None
        """
        self.coils.append(Coil(rin, rout, ymin, count_turns)) # coils is a list of Coil class objects

    def mesh(self, ny):
        """
        divide each tape into ny element evenly
        :param ny: number of elements in each tape
        :return: None
        
        posi_x is ny by m matrix where m is the number of turns of all Coils. 
            All elements in the same column have the same value being posi_x_turn=the x position of the turn
        posi_y is ny by m matrix where m is the number of turns of all Coils. 
            Each column represents a turn. 
            The elements of a column is the y values of the ny elements of a tape.
            All turns in a Coil have the same ny y-values.
        
        count_ele is the number of elements of all Coils, = ny*m 
        """
        print(time.asctime(time.localtime(time.time())) + ': Start meshing.')
        self.ny = ny
        self.dy = self.tape_type.a / ny #self.tape_type = class YBCOtape, a is the width
        self.dx = self.tape_type.b #b is tape thickness
        self.count_turns = sum([c.count_turns for c in self.coils]) #total number of turns of all Coil in the list 'coils', i.e. N_t in paper
        
        self.posi_x = np.empty(0)
        self.posi_y = np.empty(0)
        for coil in self.coils: #for each Coil object in "coils"
            # the y position of each turn in each Coil is the same. We want to divide the 
            # Coil in the y direction into ny elements. The positions are:
            # ymin + (0.5, 1.5, ..., ny-0.5)/ny*tape_width
            yk_new = (np.arange(0, ny) + 0.5) * self.tape_type.a / ny + coil.ymin
            if coil.count_turns == 1:
                dis_turn_x = 0
            else:
                dis_turn_x = (coil.rout - coil.rin) / (coil.count_turns - 1) #the distance between adjacent turns in a coil
            for turn_num in range(coil.count_turns): 
                posi_x_turn = coil.rin + turn_num * dis_turn_x
                # add the x and y positions of each turn 
                # all elements in the same turn has the same x coordinate. Each column in posi_x represents a turn
                self.posi_x = np.hstack((self.posi_x, np.ones(ny) * posi_x_turn))
                # the y positions of elements in a turn for all turns in a Coil is the same
                self.posi_y = np.hstack((self.posi_y, yk_new.copy()))
        
        self.count_ele = len(self.posi_x)
        
        #self.posi_x and self.posi_y are 1D arrays
        print(time.asctime(time.localtime(time.time())) + ': Finish meshing.')

    def set_time(self, save_time, timestep_min, timestep_max, timestep_type, j_limit=1.1):
        """
        :param save_time: times to save results
        :param timestep_min: minimum steptime
        :param timestep_max:  maximum step time
        :param timestep_type: 1: auto steptime between timestep_min and timestep_max 0: dt=timestep_min
        :param j_limit: Time step criterion used when timestep_type=1
        :return: None
        """
        self.save_time = save_time
        self.timestep_min = timestep_min
        self.timestep_max = timestep_max
        self.timestep_type = timestep_type
        self.j_limit = j_limit
        
        # plot transport current
        t_vec = np.linspace(0,self.save_time[-1],1000)
        I_vec = self._cal_i(t_vec)
        dI_vec = self._cal_di(t_vec)
        plt.plot(t_vec,I_vec)
        plt.show()
        plt.plot(t_vec,dI_vec)
        plt.show()

    def set_A(self, Bmax, Bfreq, A_type, rm_type_A,phi=0):
        """
        define the magnetic vector potential of the applied field. 
        A_type = type of external field. 
            if =1, external field B is sinusoidal in time and in the x direction only. Given in _cal_A
            if =7, external field is read from external txt file(s) (dAdt and dLapAdt)
        
        rm_type_A sets the ramp type if A_type =7.
        
        The following parameters are relevant for A_type=1 only
        Bmax  = field amplitude for B in x. 
        Bfreq = field frequency
        phi = phase angle

        :return: None
        """
        self.Bax = Bmax #unit: Tesla
        self.Aa_fre = Bfreq #unit: Hz
        self.Aa_phi = phi
        self.A_type = A_type
        self.rm_type_A = rm_type_A
        
        if self.A_type == 7:
            #load external files (s): _load_ext_files loads dAdt; _load_ext_files_withLap loads A, dAdt, dLapAdt
            self._load_ext_files()
    
    
    def _load_ext_files(self):
        """
        Load dAdt, and set up interpolation function.
        Assume the file has coordinates that are regularly spaced (dx = dy).
        The external file is generated by COMSOL.
        Manually change the time period covered by data in the file.
        """
        dAdt_array = np.loadtxt('dAdt values.txt',skiprows=9,delimiter=",")
        
        # subtract the mean to centre the positions
        dAdt_array[:,0] = dAdt_array[:,0]-np.mean(dAdt_array[:,0]);
        dAdt_array[:,1] = dAdt_array[:,1]-np.mean(dAdt_array[:,1]);

        
        # In COMSOL, the tapes have width along x axis and thickness along y axis; 
        # in Python program, coil is on 2D plane, width of the tape runs along
        # y axis, thickness of tape in x axis.
        dAdt_array[:,1] = np.round(dAdt_array[:,1],8) #truncate to 8dp
        x_data = dAdt_array[:,1];
        x_data = np.unique(x_data);
        x_data = np.sort(x_data)
        
        dAdt_array[:,0] = np.round(dAdt_array[:,0],8) #truncate to 8dp
        y_data = dAdt_array[:,0];
        y_data = np.unique(y_data);
        y_data = np.sort(y_data) #sort from smallest to biggest
        
        # Specify the times covered by data in the file.
        t_data2 = np.linspace(1/44/100,2/44,200)
        t_data1 = np.linspace(0,1/44/100,10, endpoint=False)
        t_data = np.concatenate((t_data1,t_data2))
        
        dAdt_data = np.zeros((len(x_data),len(y_data),len(t_data)));
        
        for i in range(len(x_data)):
            for j in range(len(y_data)):
                xvalue = x_data[i];
                yvalue =  y_data[j];
                # obatin the values of dAdt of all times at position (xvalue,yvalue)
                # i.e. the row where first column and second column are yvalue and
                # xvalue respectively
                rowidx1 = dAdt_array[:,1]==xvalue; #true if value = xvalue
                rowidx2 = dAdt_array[:,0]==yvalue; #true if value = yvalue
                dAdt_data[i,j,:] = dAdt_array[rowidx1*rowidx2,2:];
                
        # set up interpolation function        
        self.interp_dAdt = RegularGridInterpolator((x_data,y_data,t_data),dAdt_data);
        
        """usage
        pt = np.array([[xq1,yq1,tq1],[xq2,yq2,tq2] ])
        
        dAdtq = interp_dAdt(pt)
        """

    def _load_ext_files_withLap(self):
        """
        _load_ext_files but also load A and Lap A values.
        """
        dAdt_array = np.loadtxt('dAdt values.txt',skiprows=9,delimiter=",")
        A_array = np.loadtxt('A values.txt',skiprows=9,delimiter=",")
        dLapAdt_array = np.loadtxt('dLapAdt values.txt',skiprows=9,delimiter=",")
        
        # subtract the mean to centre the positions
        dAdt_array[:,0] = dAdt_array[:,0]-np.mean(dAdt_array[:,0]);
        dAdt_array[:,1] = dAdt_array[:,1]-np.mean(dAdt_array[:,1]);
        A_array[:,0] = A_array[:,0]-np.mean(A_array[:,0]);
        A_array[:,1] = A_array[:,1]-np.mean(A_array[:,1]);
        dLapAdt_array[:,0] = dLapAdt_array[:,0]-np.mean(dLapAdt_array[:,0]);
        dLapAdt_array[:,1] = dLapAdt_array[:,1]-np.mean(dLapAdt_array[:,1]);
        
        # Assume the files have the same x, y, and t values
        # In COMSOL, the tapes has width along the x axis and thickness along y axis; 
        # in Python the program, coil is on 2D plane, width of the tape runs along
        # y axis, thickness of tape in x
        dAdt_array[:,1]=np.round(dAdt_array[:,1],8) #truncate to 8dp
        A_array[:,1]=np.round(A_array[:,1],8) #truncate to 8dp
        dLapAdt_array[:,1]=np.round(dLapAdt_array[:,1],8) #truncate to 8dp
        x_data = dAdt_array[:,1];
        x_data = np.unique(x_data);
        x_data = np.sort(x_data)
        
        dAdt_array[:,0]=np.round(dAdt_array[:,0],8) #truncate to 8dp
        A_array[:,0]=np.round(A_array[:,0],8) #truncate to 8dp
        dLapAdt_array[:,0]=np.round(dLapAdt_array[:,0],8) #truncate to 8dp
        y_data = dAdt_array[:,0];
        y_data = np.unique(y_data);
        y_data = np.sort(y_data) #sort from smallest to biggest
    
        # Specify the times covered by data in the file.
        t_data2 = np.linspace(1/44/100,2/44,200)
        t_data1 = np.linspace(0,1/44/100,10, endpoint=False)
        t_data = np.concatenate((t_data1,t_data2))
        
        dAdt_data = np.zeros((len(x_data),len(y_data),len(t_data)));
        A_data = np.zeros((len(x_data),len(y_data),len(t_data)));
        dLapAdt_data = np.zeros((len(x_data),len(y_data),len(t_data)));
        
        for i in range(len(x_data)):
            for j in range(len(y_data)):
                xvalue = x_data[i];
                yvalue =  y_data[j];
                # obatin the values of dAdt of all times at position (xvalue,yvalue)
                # i.e. the row where first column and second column are yvalue and
                # xvalue respectively
                rowidx1 = dAdt_array[:,1]==xvalue; #true if value = xvalue
                rowidx2 = dAdt_array[:,0]==yvalue;#true if value = yvalue
                dAdt_data[i,j,:] = dAdt_array[rowidx1*rowidx2,2:];
                
                rowidx1 = A_array[:,1]==xvalue; #true if value = xvalue
                rowidx2 = A_array[:,0]==yvalue;#true if value = yvalue
                A_data[i,j,:] = A_array[rowidx1*rowidx2,2:];
                
                rowidx1 = dLapAdt_array[:,1]==xvalue; #true if value = xvalue
                rowidx2 = dLapAdt_array[:,0]==yvalue;#true if value = yvalue
                dLapAdt_data[i,j,:] = dLapAdt_array[rowidx1*rowidx2,2:];
        
                
        # set up interpolation functions
        self.interp_dAdt = RegularGridInterpolator((x_data,y_data,t_data),dAdt_data);
        self.interp_A = RegularGridInterpolator((x_data,y_data,t_data),A_data);
        self.interp_dLapAdt = RegularGridInterpolator((x_data,y_data,t_data),dLapAdt_data);
        
        """usage
        pt = np.array([[xq1,yq1,tq1],[xq2,yq2,tq2] ])
        
        dAdtq = interp_dAdt(pt)
        Aq = interp_A(pt)
        dLapAdtq = interp_dLapAdt(pt)
        """
        
    def _cal_A(self, x, y, t):
        """
        calculate the magnetic vector potential A of the applied field at different time and positions        
        :param time1: array of times
        x, y: position vectors
        :return: array of magnetic fields
        """
        print(np.shape(x),np.shape(y),np.shape(t))
        if self.A_type ==1:
            # for spatially uniform field, applied in x direction only
            return y*self.Bax*np.sin(2*np.pi*self.Aa_fre*t)
        elif self.A_type ==7:
            # interpolated value(s) using data from imported file
            pt = np.concatenate((x,y),axis=1);
            pt = np.concatenate((pt,t*np.ones((np.shape(x)))),axis=1);
            ans = self.interp_A(pt)
           
            return ans[..., np.newaxis]  

    def _cal_dAdt(self, x, y, t):
        """
        calculate dA/dt in different time and positions
        :param time1: array of times
        x, y: position vectors
        :return: array of magnetic fields gradient
        """
        if self.A_type ==1:
            # for spatially uniform field
            return y*2*np.pi*self.Aa_fre*self.Bax*np.cos(2*np.pi*self.Aa_fre*t)
        elif self.A_type ==7:
            # interpolated value(s) using data from imported file
            pt = np.concatenate((x,y),axis=1);
            pt = np.concatenate((pt,t*np.ones((np.shape(x)))),axis=1);
            ans = self.interp_dAdt(pt) * self._rm_A(10*44*t)
            return ans[..., np.newaxis]  
    
    def _cal_dLapAdt(self, x, y, t):
        """
        calculate d(del*del*A)/dt in different time and positions
        t= array of times
        x, y= position vectors
        :return: array of magnetic fields gradient
        """
        if self.A_type ==1:
            # for spatially uniform field
            return np.zeros(np.shape(y))
        elif self.A_type ==7:
            # interpolated value(s) using data from imported file
            pt = np.concatenate((x,y),axis=1);
            pt = np.concatenate((pt,t*np.ones((np.shape(x)))),axis=1);
            ans = self.interp_dLapAdt(pt) * self._rm_A(10*44*t)
            return ans[..., np.newaxis]  
          
    def set_current(self, Imax, Ia_fre,Ia_phi=0,S = 1, rm_type_I=0):
        """
        define the applied sinusoidal current
        :param Ia: current amplitude
        :param fre: current frequency
        :return: None
        """
        self.Ia = Imax #unit: A
        self.Ia_fre = Ia_fre #unit:Hz
        self.Ia_phi = Ia_phi
        self.S = S
        self.rm_type_I = rm_type_I
        

    def _cal_i(self, time1):
        """
        calculate the applied current Iset at different time
        :param time1: array of times
        :return:  array of current
        """
        if self.rm_type_I !=0:
            ans = self.S*self.Ia*np.sin(2*np.pi*self.Ia_fre*time1 +self.Ia_phi)* self._rm_I(10*self.Ia_fre*time1);
        else:
            ans = self.S*self.Ia*np.sin(2*np.pi*self.Ia_fre*time1 +self.Ia_phi)
        return ans

    def _cal_di(self, time1):
        """
        calculate the derivative dIset/dt of applied current with respect to time at different time
        :param time1: array of times
        :return:  array of dIset/dt
        """
        ans1 = 2*np.pi*self.Ia_fre *self.S*self.Ia *np.cos(2*np.pi*self.Ia_fre*time1 + self.Ia_phi)
       
        if self.rm_type_I !=0:
            ans = self.S*self.Ia*np.sin(2*np.pi*self.Ia_fre*time1 +self.Ia_phi)* 10*self.Ia_fre*self._drm_I(10*self.Ia_fre*time1) \
                + ans1*self._rm_I(10*self.Ia_fre*time1)
        else:
            ans = ans1
        return ans
    
    def run(self, build_matrix=True, cal_jcb=False): 
        """
        solve for the current density distribution
        :param build_matrix: True: rebuild the matrixes (usually set as True)
        :param cal_jcb: True: Considering Jc(B) characteristics of the tape
        :return: None
        """
        print(time.asctime(time.localtime(time.time())) + ': Start solving.')
        miu0 = 4 * np.pi * 1e-7
        # List to store current density distribution
        self.jt = [None] * len(self.save_time) # ceate a list with len(self.save_time) none-type objects. 
        self.dj = [None] * len(self.save_time)
        # List to store critical current density distribution
        self.jc = [None] * len(self.save_time)  
        
        dx = self.dx  # element size
        dy = self.dy  # element size

        Ec = 1e-4
        
        posi_x = np.reshape(self.posi_x, (self.count_ele, 1))
        posi_y = np.reshape(self.posi_y, (self.count_ele, 1)) 

        # Calculate the matrices
        if isinstance(self.L, type(None)) or build_matrix:
            print(time.asctime(time.localtime(time.time())) + ': Start building matrices. ')
            print(time.asctime(time.localtime(time.time())) + ': Cartesian coordinate system.')
            self._build_matrix_car()
        else:
            print(time.asctime(time.localtime(time.time())) + ': Using last matrices.')

        # Initial data structures
        time_run = 0    # model time
        self.step = 0   # model step
        self.save_time_step.append(self.step)  # Array to record step numbers (integers) corresponding to times when results are saved
        self.time_point_no = 1  # time_point_no is index of elements in time_save (the times to save result)
        self.dta = []  # Array to record length of each step time used in solving
        self.time_array = []  # Array to record the time for each step
        self.power2 = np.empty((self.count_turns,0))  # Matrix to record the AC loss in each step
        
        lastjt = np.zeros((self.count_ele, 1))  # current density J in last step. Initially, at time t=0, lastjt is zero
        
        self.jc[0] = np.ones((self.count_ele, 1)) * self.tape_type.jc  # Critical current density at step [0]
        self.jt[0] = lastjt.copy()  # current density at the 0th time instant to save result, assume current density to be zero at time t=0
        reach_save = False  # switch to record
        
      
        # Main solving loop
        while self.time_point_no < len(self.save_time):
            self.time_array.append(time_run)  # record the time at this iteration
            
            ## 1. Calculate critical current density of each point and the electric field E 
            # jc constant in this simulation
            jc = self.tape_type.jc
            
            # E is the electric field at each point, and is based on the E-J power law, using the J in the last time step
            E = np.abs(Ec * (lastjt / jc)**self.tape_type.nvalue) * np.sign(lastjt) 
            
            if np.isnan(E[0]):
                # Record the J when the calculation does not converge
                self.breakj = lastjt
                print ('**calculation failed to converge at time %f at step %i !! \n' %(time_run,self.step))
                break
            
            ## 2. Calculate instantaneous ac loss
            # power_temp is a self.count_turns by 1 vector storing the power loss per unit vol at each turn.
            # done by first multiplying E and J at each element, then reshape into self.count_turns by self.ny matrix, and then sum along each row
            power_temp = (E * lastjt).reshape((self.count_turns, self.ny)).sum(axis=1)  # calculate the AC loss
            self.power2 = np.hstack((self.power2, np.transpose([power_temp])))
            
            
            ## 3. Calculate dj = J_dot = derivative of J wrt time
            ## 3a. Calculate the first part of dj
            jtdb = np.matmul(self.Qij_inv, E + self._cal_dAdt(posi_x,posi_y,time_run))
            
            ## 3b. Calculate the third part of dj, uncomment when Laplacian of A_a is taken into account
            # jt3 = -1/miu0*self._cal_dLapAdt(posi_x,posi_y,time_run)

            ## 3c. Calculate second part of dj
            # dx dy located differently compared to paper, but pphipz should remain unchanged. 
            # pphipz is -partial phi by partial z
            P = self._cal_di(time_run)/dx/dy*np.ones((self.count_turns, 1))  -np.matmul(self.L, jtdb) #  -np.matmul(self.L, jt3) #uncomment last term if want to include Laplacian of A_a in calculation 
      
            pphipz = np.matmul(self.LM_inv, P)

            jtdc = np.matmul(self.M, pphipz)
            
            dj = jtdb + jtdc #+jt3 #uncomment last term if want to include Laplacian of A_a in calculation 
            
            if self.step ==0:
                self.jt[0] = dj.copy() #dj at time 0

            
            ## 4. Calculate J based on dj and dt. Advance J by J(t) = J(t-1) + dj*dt
            
            # 4a. Calculate time step dt to take
            # the default dt is dts calculated as follows, using adaptive time stepping if timestep_type==1
            if self.timestep_type == 1:
                # auto dt
                s1 = dj > 0
                s2 = dj < 0
                dts = self.timestep_max
                if s1.max():
                    dts = min(self.timestep_max, abs((self.j_limit * jc - lastjt)[s1] / dj[s1]).min())
                if s2.max():
                    dts = min(dts, abs((-self.j_limit * jc - lastjt)[s2] / dj[s2]).min())
                dts = max(dts, self.timestep_min)
            else:
                # fixed dt
                dts = self.timestep_min
            
            # However, dt modified from dts if the current time (time_run) at which J_dot 
            # is calculated + dts will go pass the next fixed time point (as elements in array save_time)
            # that we want to save our result. In such case, we only advance by dt such that we reach the
            # next set time point and not overshoot
            if dts >= self.save_time[self.time_point_no] - time_run:
                # Reach the save point
                reach_save = True
                dt = self.save_time[self.time_point_no] - time_run
            else:
                dt = dts
            
            # 4b. Record dt, update time_run (the current time for next time step), J, and the step number
            self.dta.append(dt)  # Record dt
            time_run += dt  # update time. time_run is the actual time, t, of the result
            lastjt += dj * dt  # update current. lastjt is the current density (of all points) at the last time step
            self.step += 1  # update step. Step number goes from 0,1,2,...
            
            if reach_save:
                # Record jt jc
                self.dj[self.time_point_no] = dj.copy()
                self.jt[self.time_point_no] = lastjt.copy() #save the current density values in list jt
                self.jc[self.time_point_no] = jc.copy() #save the critical current density values in list jc

                if np.shape(self.save_time)[0]>500: #if there are lots of (more than 500) time points at which we want to save results, we only print progress every 100 points 
                    if self.time_point_no%100==0:
                        print(time.asctime(time.localtime(time.time())) + ': Result at time = %5.4f is recorded.' % time_run)
                else:
                    print(time.asctime(time.localtime(time.time())) + ': Result at time = %5.4f is recorded.' % time_run)
                self.time_point_no += 1 #go to the next time point. Started with 1 initially
                #save_time_step stores the step numbers when the set time points to save results are reached
                self.save_time_step.append(self.step) 
                reach_save = False
            
            
        print(time.asctime(time.localtime(time.time())) + ': Finish solving.')

        
    def _build_matrix_car(self):
        """
        Build the matrixes used in Cartesian coordinate system
        :return: None
        
        Matrices built include:
            Qij (but not stored as an attribute of AClossFT2D class object, i.e. no self.Qij)
            Qij_inv 
            M
            L
            LM_inv
        """
        miu0 = 4 * np.pi * 1e-7
        dx = self.dx
        dy = self.dy
        # r1_x is Ne × Ne matrix whose rows are just posi_x
        # r2_x is Ne × Ne matrix whose columns are just posi_x
        r1_x, r2_x = np.meshgrid(self.posi_x, self.posi_x) #self.posi_x and self.posi_y are 1D arrays
        r1_y, r2_y = np.meshgrid(self.posi_y, self.posi_y)

        # self.L is a matrix containing the relation between the element number and its turn number
        # L is N_t by N_e, where N_t = total number of turns summed over all Coils, N_e = N_t*ny = total number of elements meshed
        # Each column thus represents an element, and in a given column, the i^th row is 1 if the element belongs to turn number i (turn number = 0, 1, ..., N_t-1)
        self.L = np.zeros((self.count_turns, self.count_ele)) 
        for i in range(self.count_turns): #for each turn number
            self.L[i, i * self.ny: (i + 1) * self.ny] = np.ones(self.ny)
        
        # Matrix Qij = Q in paper
        # Qij is an N_e by N_e matrix, N_e = number of elements
        # Let a1,b1,c1,d1 be Ne × Ne matrices whose rows are the left x, right x, bottom y and top y coordinates of each element. Rows repeated N_e times in each matrix. 
        a1 = r1_x - dx/2
        b1 = r1_x + dx/2     
        c1 = r1_y - dy/2               
        d1 = r1_y + dy/2

        u2= r2_x - b1 # x_i - b_j
        u1 = r2_x - a1 # x_i - a_j
        v2 = r2_y - d1# y_i - d_j
        v1 = r2_y - c1 # y_i - c_j      

        Qij = -1/4/np.pi*(u2*v2*(np.log(u2**2+v2**2)-3)+u2**2*np.arctan(v2/u2)+v2**2*np.arctan(u2/v2) \
                          -u1*v2*(np.log(u1**2+v2**2)-3)-u1**2*np.arctan(v2/u1)-v2**2*np.arctan(u1/v2) \
                          -u2*v1*(np.log(u2**2+v1**2)-3)-u2**2*np.arctan(v1/u2)-v1**2*np.arctan(u2/v1) \
                          +u1*v1*(np.log(u1**2+v1**2)-3)+u1**2*np.arctan(v1/u1)+v1**2*np.arctan(u1/v1) )/dx/dy
        

        ## calculate Qij_inv, M, LM_inv
        del r1_x, r2_x, r1_y, r2_y
        
        self.Qij_inv = -1 / miu0 / dx / dy * np.linalg.inv(Qij)

        del Qij
        
        # LM_inv is M^-1 in paper. The M below does not correspond to the M in paper.
        self.M = -np.matmul(self.Qij_inv, self.L.transpose()) 

        LM = np.matmul(self.L, self.M) 
   
        self.LM_inv = np.linalg.inv(LM)

        print(time.asctime(time.localtime(time.time())) + ': Finish building matrices.')

 
    
    def plot_resultj(self, timestep_n0):
        """
        plot the normalized current density distribution at timestep_n0
        :param timestep_n0: we want to show the result at the timestep_n0^th time step in save_time (the time instants to save results)
        """
        if timestep_n0 >= len(self.save_time):
            print(time.asctime(time.localtime(time.time())) + ': Result at timestep %i is not saved.' % timestep_n0)
            return
        #j is the normalized current density

        j = (self.jt[timestep_n0] / self.jc[timestep_n0]).reshape((self.count_turns, self.ny))
        
        x = self.posi_x.reshape((self.count_turns, self.ny))
        y = self.posi_y.reshape((self.count_turns, self.ny))
        
        f = plt.figure()

        ax = f.add_subplot(111)
        plotme = ax.contourf
        
        k0 = 0
        ax.set_xlabel('x (m)')
        ax.set_ylabel('y (m)')
        time_shown = self.save_time[timestep_n0]
        ax.set_title('J/Jc at time %f' % time_shown)
        
        clev = np.arange(-1.5,1.5,.001)  # set range of colormap 
        
        for i in range(len(self.coils)):
            k1 = k0 + self.coils[i].count_turns
            surf = plotme(x[k0: k1, :], y[k0: k1, :], j[k0: k1, :], clev,cmap=cm.coolwarm)
            k0 = k1
        
        cb = f.colorbar(surf, ticks=[-1, -0.5, 0, 0.5, 1])
        

    def post_cal_power(self, time_point_no0, time_point_no1):
        """
        Calculate the ac loss between time_point_no0 and time_point_no1
        :return: Average ac loss, and two arrays: runtime_array,loss_array, 
        which give the times and instantaneous losses at those times respectively for plotting
        """

        # save_time_step1 is the step number corresponding to time represented by time_point_no0
        save_time_step1 = self.save_time_step[time_point_no0] 
        save_time_step2 = self.save_time_step[time_point_no1]
        power_avg = np.zeros(self.count_turns)
        for i in range(self.count_turns):
            power2 = self.power2[i,save_time_step1:save_time_step2] * self.dx * self.dy
   
            #dt is the length of each time step 
            dt = np.array(self.dta[save_time_step1:save_time_step2])
            
            # power_avg stores the average power loss for each turn i during the time specified
            power_avg[i] = (power2 * dt).sum() / dt.sum()
        
        ## plot the (loss of all coils) over time
        f = plt.figure()
        axe = f.add_subplot(111)
        axe.set_xlabel('Time (s)')
        axe.set_ylabel('Instantaneous loss  (W/m)')
        #axe.set_ylim([0,8])

        
        save_time_step1 = 0 # want to plot loss from time=0
        runtime=self.save_time[save_time_step1]
        
        axe.set_xlim([runtime,self.save_time[time_point_no1]])
        
        i=0
        runtime_array = np.zeros(save_time_step2-save_time_step1) #store the actual time (runtime) corresponding to each time step
        loss_array = np.zeros(save_time_step2-save_time_step1) #store the loss at each time step
        
        #[self.power2]_ij stores the instantaneous loss of turn i at time step j
        for step in range(save_time_step1,save_time_step2):
            runtime_array[i]=runtime
            loss_array[i] = np.sum(self.power2[:,step])* self.dx * self.dy
            
            i = i+1
            runtime = runtime+self.dta[step]
        
        axe.plot(runtime_array, loss_array, marker = ",", linestyle = 'None',)

        axe.grid(which='both')
        axe.set_title('Instantaneous loss over time, \n ny = %d' %self.ny)
        

        ## plot the average loss for each turn, each coil is one line
        f = plt.figure()
        axe = f.add_subplot(111)
        startnum = 0
        for i in range(len(self.coils)):
            endnum = startnum + self.coils[i].count_turns
            axe.semilogy(np.arange(1, self.coils[i].count_turns + 1), power_avg[startnum: endnum], marker = "x")
            startnum = endnum
        axe.set_xlabel('Turn number')
        axe.set_ylabel('Average loss (W/m)')
        axe.grid(which='both')
        axe.set_title('Average loss for each turn organised by coils, \n ny = %d' %self.ny)

        plt.show()
        
        power_avg_all_coils = power_avg.sum() #the average instantaneous power loss over the time specified of all the coils
        print(time.asctime(time.localtime(time.time())) + ': Total average loss: %5.4f W/m.' % (power_avg_all_coils))
        
        return power_avg_all_coils, runtime_array,loss_array 

    def _find_ticks(self,max_value, position_of_zero):
        """find tick positions that go up either in 5 or 2 times 10**x
        for plotting
        """
        power = np.floor(np.log10(max_value))

        if position_of_zero =='bottom':
            #max_value*10**(-power) should be between 0 and 10
            if max_value*10**(-power)<4:
                interval = 5*10**(power-1)
            elif max_value*10**(-power)<8:
                interval = 1*10**(power)
            else:
                interval = 2*10**(power)
            
            stop = np.floor(np.divide(max_value,interval))*interval
            ticks=np.linspace(0,stop,int(stop/interval+1))
    
        elif position_of_zero =='centre':
            if max_value*10**(-power)<=2:
                interval = 5*10**(power-1)
            elif max_value*10**(-power)<6:
                interval = 1*10**(power)
            else:
                interval = 2*10**(power)
        
            stop = np.floor(np.divide(max_value,interval))*interval
            ticks=np.linspace(-stop,stop,int(2*stop/interval+1))
    
        return ticks
        
   
    def plot_cal_dAdt(self, timestep_n0):
        """
        plot dAdt at different times at all elements' positions

        """
        x = self.posi_x.reshape((self.count_turns, self.ny))
        y = self.posi_y.reshape((self.count_turns, self.ny))
        t = self.save_time[timestep_n0]
        dAdt = self._cal_dAdt(x, y, t)

        f = plt.figure()
        ax = f.add_subplot(111)
        ax.set_xlabel('x (m)')
        ax.set_ylabel('y (m)')
        time_shown = self.save_time[timestep_n0]
        ax.set_title('dAdt at time %f' % time_shown)
        k0 = 0
        
        clev = np.arange(dAdt.min(),dAdt.max(),(dAdt.max()-dAdt.min())/100)  # set range of colormap 
        
        for i in range(len(self.coils)):
            k1 = k0 + self.coils[i].count_turns
            surf = ax.contourf(x[k0: k1, :], y[k0: k1, :], dAdt[k0: k1, :], clev, cmap=cm.coolwarm)
            k0 = k1
        maxvalue = dAdt.max()

        cb = f.colorbar(surf, ticks=self._find_ticks(dAdt.max(),'centre'))

    def plot_cal_dLapAdt(self, timestep_n0):
        """
        plot dLapAdt at different times at all elements' positions
        """
        x = self.posi_x.reshape((self.count_turns, self.ny))
        y = self.posi_y.reshape((self.count_turns, self.ny))
        t = self.save_time[timestep_n0]
        dLapAdt = self._cal_dLapAdt(x, y, t)
    
        f = plt.figure()
        ax = f.add_subplot(111)
        ax.set_xlabel('x (m)')
        ax.set_ylabel('y (m)')
        time_shown = self.save_time[timestep_n0]
        ax.set_title('dLapAdt at time %f' % time_shown)
        k0 = 0
         
        clev = np.arange(dLapAdt.min(),dLapAdt.max(),(dLapAdt.max()-dLapAdt.min())/100)  # set range of colormap 
         
        for i in range(len(self.coils)):
            k1 = k0 + self.coils[i].count_turns
            surf = ax.contourf(x[k0: k1, :], y[k0: k1, :], dLapAdt[k0: k1, :], clev, cmap=cm.coolwarm)
            k0 = k1
        maxvalue = dLapAdt.max()
    
        cb = f.colorbar(surf, ticks=self._find_ticks(dLapAdt.max(),'centre'))
         

    def save_result(self, filename='./ACloss_YBCO2D.pkl'):
        """
        save the calculated results
        :param filename: filename
        :return: None
        """
        with open(filename, 'wb') as f:
            pickle.dump(self, f)

    #@staticmethod
    def load_result(self, filename):
        """
        load saved result
        :param filename: filename
        :return: (AClossFT2D) loaded results
        """
        with open(filename, 'rb') as f:
            data = pickle.load(f)
        return data








