"""
This program can be used with ACloss_class.py to calculate the current distribution
of REBCO tapes under non-uniform external field (imported from external file) 
whilst carrying transport current. As an example, loss in 20 tapes is calculated below.

Reference paper:
C. C. T. Chow, F. Grilli, and K. T. Chau, "Numerical modelling of HTS tapes 
under arbitrary external field and transport current via integral method: 
review and application to electrical machines," Supercond. Sci. Technol., 
2023. [Online]. Available: http://iopscience.iop.org/article/10.1088/1361-6668/ace701.

Author: Calvin C. T. Chow (ctcchow@eee.hku.hk)

Version history:
v1.0: October 2023 First published version. In Python 3.8.

Based on code by Dr. Lingfeng Lai dated 2020/10/30 available at https://www.htsmodelling.com/?wpdmpro=j_model

This program is free software: you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by the Free Software 
Foundation, either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT 
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with 
this program. If not, see <https://www.gnu.org/licenses/>.
"""


from ACloss_class import YBCOtape, AClossFT2D

import numpy as np
import matplotlib.pyplot as plt
import time
from decimal import Decimal
import platform

computer_info = platform.processor() #check current computing platform for reporting purposes

### 1. set geometry of problem
# define the YBCO tape
tape = YBCOtape(a=4e-3, b=2e-6, nvalue=30, jc0=1.149e10)

# Set solver
solver = AClossFT2D(tape) 

# adding a coil (left boundary, right boundary, bottom y position, number of turns)
bound = 0.20384e-3*52/20*19/2
solver.add_coil(-bound, bound, -0.002, 20)


### 2.  set external conditions
# set external magnetic vector potential (A)
Bmax = 0 # [T] #only relevant for A_type=1
Bfreq = 20 #[Hz] #only relevant for A_type=1
A_type = 7 # =1 for uniform external field, =7 for external field to be taken from imported file
rm_type_A = 0; #there is already a ramp in COMSOL simulation that generated the externally imported file, so no further ramp needed
solver.set_A(Bmax, Bfreq, A_type,rm_type_A)

## set the transport current
Imax = 73.54 #[A]
Ia_fre = 44 #[Hz]
alpha = 0 #offset used in machine
Ia_phi= np.pi/2 + alpha # default is sine function, but COMSOL has cosine
S = -1
rm_type_I = 2; #set ramp in transport current
solver.set_current(Imax, Ia_fre, Ia_phi,S, rm_type_I)



### 3.  set solver parameters
# divide each tape into ny elements along the y direction equally.
ny=80
solver.mesh(ny) 
# Set the simulated times.
# Simulate over two periods
# Results (instantaneous loss, current density etc.) will be stored at exact time instants specified below
save_time = np.linspace(0,2/44,2001)
   
# minimum and maximum time steps
timestep_min = 1e-9
timestep_max = 1e-7 

solver.set_time(save_time, timestep_min, timestep_max, 1)

### plot background magnetic vector potential
# solver.plot_cal_dLapAdt(8) #argument: time_point_no, i.e. the index in save_time, corresponding to the time you want to plot the result for
# solver.plot_cal_dAdt(8)

### 4. solve
t = time.time()
solver.run()
elapsed = time.time() - t
print('elapsed time (s)=', elapsed)


#save whatever result, may not have converged
save_result = 1
if save_result==1:
    timestr = time.asctime(time.localtime(time.time())).replace(":","")
    filename = './results/' + timestr + '.pkl'
    filename_time = './results/' + timestr + 'time_taken'
    filename_computer = './results/' + timestr + 'computer.txt'
    solver.save_result(filename)


#plot results at specified frames
frames = [101,201] #enter in brackets the time_point_no
for frame in frames:
     solver.plot_resultj(frame) #normalized current density of all turns


#calculate the average ac loss for the second period simulated 
power_avg, runtime_array,loss_array = solver.post_cal_power(1000,2000) #enter indices in save_time corresponding to the second period simulated
np.savetxt('python_runtime_array.txt', runtime_array, delimiter=',')
np.savetxt('python_loss_array.txt', loss_array, delimiter=',')

# save results
if save_result==1:
    solver.save_result(filename)
    with open(filename_computer, 'w') as f:
        f.write(computer_info+'. time_taken = '+str(elapsed))

    




